package ch.hevs.kart;

/**
 * Interface to implement if an object is interested in basic kart status changes. An object 
 * implementing this interface can be added using the Kart's method addKartListener(). The object
 * will get notified about changes of the most important parameters of the Kart.
 * 
 * Note that not all changes in the registers of the kart are announced by this listener interface. 
 * If an object needs more details about all the registers of the Kart, it should implement the 
 * {@link KartStatusRegisterListener} interface and subscribe himself using the method
 * addStatusRegisterListener().
 * 
 * @author Nils Chatton
 * @author Christopher Metrailler (mei@hevs.ch)
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.1 - 2015
 */
@SuppressWarnings("UnusedParameters")
public interface KartListener {
	/**
	 * Called by the card whenever the actual steering position has changed.
	 * 
	 * This method is part of the KartListener interface and all listeners get notified when the 
	 * value of the actual steering position has changed.
	 * 
	 * @param kart					The kart object.
	 * @param absolutePosition		The actual position from 0 to max steps (depends kart model!).
	 * @param normalizedPosition	The actual position from -1 (left) to 1 (right).
	 */
	void steeringPositionChanged( Kart kart, int absolutePosition, float normalizedPosition);
	
	/**
	 * The steering system has changed its state.
	 * 
	 * This method is called when either the steering position has just reached its end position 
	 * (reached = true) or when the steering system has got a new target value and starts the 
	 * moving to this position (reached = false).
	 * 
	 * @param kart		The kart object.
	 * @param reached	True if the steering target position has reached, false otherwise.
	 */
	void steeringPositionHasReached(Kart kart, boolean reached);
	
	/**
	 * The battery level has changed.
	 * 
	 * This method is called by the kart object whenever the battery level on the kart changes.
	 * 
	 * @param kart		The kart object.
	 * @param level		The battery level from 0 (empty) to 1 (full).
	 */
	void batteryVoltageChanged(Kart kart, float level);
	
	/**
	 * The connection status to the Kart has changed.
	 * 
	 * @param kart		The kart object.
	 * @param online	True if the kart is online, false otherwise.
	 */
	void connectionStatusChanged(Kart kart, boolean online);
	
	/**
	 * The kart has a message to display, mostly this are cryptic messages just for 
	 * 		  developers...
	 * 
	 * @param kart		The kart object.
	 * @param message	The message.
	 */
	void message(Kart kart, String message);
}
