package ch.hevs.kart;

import java.security.InvalidParameterException;

/**
 * Enumeration in order to identify the different status registers of a kart.
 *
 * @see KartStatusRegisterListener
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.1 - 2015
 */
public enum KartStatusRegister {
	/**
	 * Hall sensor 1 pulse count.
	 */
	HallSensorCounter1(0),
	
	/**
	 * Hall sensor 2 pulse count.
	 */
	HallSensorCounter2(1),
	
	/**
	 * Steering FSM status: 0=busy 1=target position reached.
	 */
	SteeringStatus(2),

	/**
	 * Actual steering position.
	 */
	SteeringPosition(3),
	
	/**
	 * State of the steering end switch, 0=closed, 1=open.
	 */
	SteeringEndSwitchState(4),
	
	/**
	 * ADC value of the battery voltage level measure.
	 */
	BatteryVoltageLevel(5),
	
	/**
	 * Ultrasonic distance sensor value. Bigger values means more distance.
	 */
	DistanceSensor(6),

    /**
     * Unused register.
     */
    Unused(7),

	/**
	 * Proximity sensor 1 value.
	 */
	ProximitySensor1(8),
	
	/**
	 * Proximity sensor 2 value.
	 */
	ProximitySensor2(9),
	
	/**
	 * Proximity sensor 3 value.
	 */
	ProximitySensor3(10),
	
	/**
	 * Proximity sensor 4 value.
	 */
	ProximitySensor4(11),
	
	/**
	 * Proximity sensor 1 ambient light measure.
	 */
	AmbientLightSensor1(12),
	
	/**
	 * Proximity sensor 2 ambient light measure.
	 */
	AmbientLightSensor2(13),
	
	/**
	 * Proximity sensor 3 ambient light measure.
	 */
	AmbientLightSensor3(14),
	
	/**
	 * Proximity sensor 4 ambient light measure.
	 */
	AmbientLightSensor4(15);
		
	KartStatusRegister(final int address) {
		this.address = address;
	}
	
	/**
	 * Returns the actual address of the status register in the Kart's memory.
	 *
	 * @return Status register address.
	 */
	public int getAddress() {
		return address;
	}
	
	/**
     * Creates a status register enumerator from the given register address.
     *
     * @param address                       The address of the register in the kart's register memory.
     * @return                              The address enumerator.
     * @throws InvalidParameterException    If no register is found at the given address.
     */
	public static KartStatusRegister fromAddress(final int address)
            throws InvalidParameterException {
		for (KartStatusRegister register: values()) {
			if (address == register.getAddress())
				return register;
		}
		throw new InvalidParameterException("No KartStatusRegister with address " + address +
                " exists!");
	}

    // The total number of status registers.
    static final int COUNT = 16;

	// The address of the register.
	private final int address;
}
