package ch.hevs.kart.comm.protocol;

import android.os.Handler;
import android.os.Looper;
import android.os.Message;
import ch.hevs.kart.KartStatusRegister;

/**
 * Wrapper to ensure that the callbacks from the protocol are always called inside the 
 * 		  application's main thread (Android is very fussy about that).
 * 
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.1 - 2015
 */
public class UiThreadProtocolListenerWrapper implements ProtocolListener {
	/**
	 * Sets the actual listener wrapped by the thread wrapper.
	 * 
	 * @param listener The listener to set up.
	 */
	public void setListener(final ProtocolListener listener) {
		this.listener = listener;
	}

	// The type of the message (We send messages between threads to transfer the callbacks).
	private enum MessageType {
		RegisterValueIndication,
		StateChangeIndication,
		ErrorIndication
	}
	
	// The actual message content we send between the threads.
	private class MessageContent {
		// Creates a new message for the registerValueIndication() method.
		public MessageContent(final Protocol protocol, final KartStatusRegister register,
                              final int value, final ProtocolListener listener) {
			this.type = MessageType.RegisterValueIndication;
			this.register = register;
			this.subject = value;
			this.protocol = protocol;
			this.listener = listener;
		}
		
		// Creates a new message for the stateChangeIndication() method.
		public MessageContent(final Protocol protocol, final ProtocolState state,
                              final ProtocolListener listener) {
			this.type = MessageType.StateChangeIndication;
            this.register = KartStatusRegister.Unused;
			this.subject = state;
			this.protocol = protocol;
			this.listener = listener;
		}
		
		// Creates a new message for the errorIndication() method.
		public MessageContent(final Protocol protocol, final String error,
                              final ProtocolListener listener) {
			this.type = MessageType.ErrorIndication;
            this.register = KartStatusRegister.Unused;
			this.subject = error;
			this.protocol = protocol;
			this.listener = listener;
		}
		
		public final MessageType type;
		public final KartStatusRegister register;
		public final Object subject;
		public final Protocol protocol;
		public final ProtocolListener listener;
	}
	
	@Override
	public void registerValueIndication(final Protocol protocol, final KartStatusRegister register,
                                        final int value) {
		// If listener is empty, we have nothing to do.
		if (listener == null) return;
		
		// If we are actually in the main thread, we call the listener method directly.
		if (Looper.getMainLooper().getThread() == Thread.currentThread()) {
			listener.registerValueIndication(protocol, register, value);
		} else {
			// Otherwise we send a message to the main loop that will call the listener later.
			Message message = Message.obtain();
			message.obj = new MessageContent(protocol, register, value, listener);
			handler.sendMessage(message);
		}
	}

	@Override
	public void stateChangeIndication(final Protocol protocol, final ProtocolState state) {
		// If listener is empty, we have nothing to do.
		if (listener == null) return;
		
		// If we are actually in the main thread, we call the listener method directly.
		if (Looper.getMainLooper().getThread() == Thread.currentThread()) {
			listener.stateChangeIndication(protocol, state);
		} else {
			// Otherwise we send a message to the main loop that will call the listener later.
			Message message = Message.obtain();
			message.obj = new MessageContent(protocol, state, listener);
			handler.sendMessage(message);
		}
	}

	@Override
	public void errorIndication(final Protocol protocol, final String error) {
		// If listener is empty, we have nothing to do.
		if (listener == null) return;
		
		// If we are actually in the main thread, we call the listener method directly.
		if (Looper.getMainLooper().getThread() == Thread.currentThread()) {
			listener.errorIndication(protocol, error);
		} else {
			// Otherwise we send a message to the main loop that will call the listener later.
			Message message = Message.obtain();
			message.obj = new MessageContent(protocol, error, listener);
			handler.sendMessage(message);
		}
	}

	// The real listener wrapped by this object.
	private ProtocolListener listener;
	
	// Main thread handler that will handle the messages posted by this wrapper.
	private static class MessageHandler extends Handler {
		@Override
		public void handleMessage(Message message) {
			// Get message content and switch depending the message type.
			MessageContent content = (MessageContent)message.obj;
			switch (content.type) {
				case RegisterValueIndication:
					// Call the registerValueIndication() on the real listener.
					if (content.listener != null)
						content.listener.registerValueIndication(content.protocol, content.register,
                                (Integer)content.subject);
					break;

				case StateChangeIndication:
					// Call the stateChangeIndication() on the real listener.
					if (content.listener != null)
						content.listener.stateChangeIndication(content.protocol,
								(ProtocolState)content.subject);
					break;
				case ErrorIndication:
					// Call the errorIndication() on the real listener.
					if (content.listener != null)
						content.listener.errorIndication(content.protocol, (String)content.subject);
					break;
			}
		}
	}

	private static final Handler handler = new MessageHandler();
}
