package ch.hevs.kart.minimalistic;

import android.graphics.Color;
import android.os.Bundle;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnTouchListener;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.CompoundButton;
import android.widget.CompoundButton.OnCheckedChangeListener;
import android.widget.TextView;
import android.widget.Toast;
import ch.hevs.kart.Kart;
import ch.hevs.kart.ui.AbstractKartControlActivity;
import ch.hevs.kart.R;

/**
 * Basic kart control activity. Just offers some very basic controls for the kart.
 * 
 * The method onCreate() is called upon the creation of the activity and thus is similar to the 
 * main() method in standard Java applications. Here you can instantiate your objects and setup
 * your activity.
 * Additionally Android activities can be paused and resumed later by the operating system. For
 * example if a call is received the actual active activity is send to the background and 
 * resumed once the call has ended. In order to react to those changes, an activity can have the 
 * onResume() and onPause() methods. Use these to start/stop your activities and create or close
 * your connections.
 * The connection to the kart and the registering of the activity as listener is already handled 
 * in the base class AbstractKartControlActivity,  so you do not have to connect or disconnect 
 * to/from the kart and you do not need to call setKartLister() as this is already done for you. 
 * However it is very important to call the respective onCreate(), onResume() and onPause()
 * methods on the super class.
 * 
 * @author Nils Chatton
 * @author Christopher Metrailler (mei@hevs.ch)
 * @author Michael Clausen (clm@hevs.ch)
 * 
 * @version 2.1 - 2015
 */
public class MinimalisticKartControlActivity extends AbstractKartControlActivity
		implements OnClickListener, OnCheckedChangeListener, OnTouchListener {
	
	private View view;					// The view of the activity.
	private TextView steeringPosition;	// The text view that will show the actual steering angle.
	private TextView batteryVoltage;	// The text view that will show the battery level.
	
	
	/*** Android activity life-cycle implementation ***********************************************/
	
	/**
	 * Activity initialization.
	 * 
	 * This method is called in order to initialize the activity. It is very similar to the  
	 * main() method of a standard Java application. However in Android activities this main() is
	 * divided into two methods: onCreate() is called when the activity gets actually created and
	 * onResume() is called when the activity actually is on the device screen. So you should just
	 * create your objects and relations between them here, the rest of the code should go into 
	 * onResume(). Even it the method is named onResume() it will get called at activity start too.
	 * 
	 * @param savedInstanceState	The saved bundle state (from last run).
	 */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		// It is important to first call the constructor of the Activity base class!
		super.onCreate(savedInstanceState);
		
		// We set the given resource as the main view of the activity.
		setContentView(R.layout.minimal_layout);
		
		// Find (sub) views and store references to them in order to manipulate them...
		view = findViewById(R.id.BasicLayout);
		steeringPosition = (TextView)findViewById(R.id.PositionTextView);
		batteryVoltage = (TextView)findViewById(R.id.BatteryVoltage);
		
		// Register this activity as the click handler for faster and slower buttons.
		Button button = (Button)findViewById(R.id.FasterButton);
		button.setOnClickListener(this);
		button = (Button)findViewById(R.id.SlowerButton);
		button.setOnClickListener(this); 
		
		// Register this activity as the checked change handler for the LED checkboxes.
		CheckBox checkBox = (CheckBox)findViewById(R.id.Led1);
		checkBox.setOnCheckedChangeListener(this);
		checkBox = (CheckBox)findViewById(R.id.Led2);
		checkBox.setOnCheckedChangeListener(this);
		checkBox = (CheckBox)findViewById(R.id.Led3);
		checkBox.setOnCheckedChangeListener(this);
		checkBox = (CheckBox)findViewById(R.id.Led4);
		checkBox.setOnCheckedChangeListener(this);
		
		// Register this activity as the touch event handler for the left and right buttons.
		button = (Button)findViewById(R.id.LeftButton);
		button.setOnTouchListener(this);
		button = (Button)findViewById(R.id.RightButton);
		button.setOnTouchListener(this);
	}
	
	/**
	 * Activity is resumed.
	 * 
	 * This method is called just before the activity is presented to the user on the screen.
	 */
	@Override
	public void onResume() {
		// Call the method on the super class.
		super.onResume();
		
		// Until we actually have a connection to the kart, we set the view as not enabled.
		setEnabledForAllSubviews(view, false);
	}
	
	/**
	 * Activity is paused.
	 * 
	 * This method is called just before the activity is removed from the screen. This might be
	 * because the user closes the activity or because another activity takes the main screen.
	 */
	@Override
	@SuppressWarnings("EmptyMethod")
	public void onPause() {
		// Call the method on the super class.
		super.onPause();
	}

	/**
	 * Activity should save its state.
	 * 
	 * If Android needs more memory for other activities an activity in the background can be
	 * terminated at any time. Just before the activity is terminated, Android gives the activity
	 * the chance to save its state into a bundle. The contents of this bundle is then given to
	 * the activity the next time it will get started to the method onCreate().
	 * 
	 * @param outState	The bundle object where the state can be written to.
	 */
	@Override
	@SuppressWarnings({"NullableProblems", "EmptyMethod"})
	protected void onSaveInstanceState(Bundle outState) {
		// Call the method on the super class.
		super.onSaveInstanceState(outState);
	}


	/*** KartListener interface implementation (Kart -> Activity) *********************************/
	
	/**
	 * Called by the card whenever the actual steering position has changed.
	 * 
	 * This method is part of the KartListener interface and all listeners get notified when the 
	 * value of the actual steering position has changed.
	 * 
	 * @param kart					The kart object.
	 * @param absolutePosition		The actual position from 0 to max steps (depends kart model!).
	 * @param normalizedPosition	The actual position from -1 (left) to 1 (right).
	 */
	@Override
	public void steeringPositionChanged(Kart kart, int absolutePosition, float normalizedPosition) {
		// Update the steering wheel position text on the view.
		steeringPosition.setText(String.valueOf(absolutePosition));
	}

	/**
	 * The steering system has changed its state.
	 * 
	 * This method is called when either the steering position has just reached its end position 
	 * (reached = true) or when the steering system has got a new target value and starts the 
	 * moving to this position (reached = false).
	 * 
	 * @param kart		The kart object.
	 * @param reached	True if the steering target position has reached, false otherwise.
	 */
	@Override
	public void steeringPositionHasReached(Kart kart, boolean reached) {
		// If the position has reached, set the color to green, red otherwise.
		steeringPosition.setTextColor(reached ? Color.GREEN : Color.RED);
	}

	/**
	 * The battery level has changed.
	 * 
	 * This method is called by the kart object whenever the battery level on the kart changes.
	 * 
	 * @param kart		The kart object.
	 * @param level		The battery level from 0 (empty) to 1 (full).
	 */
	@Override
	public void batteryVoltageChanged(Kart kart, float level) {
		// Update the battery level text view.
		final String l = String.format("%.2f %%", level * 100);
		batteryVoltage.setText(l);
	}

	/**
	 * The connection status to the Kart has changed.
	 * 
	 * @param online	True if the kart is online, false otherwise.
	 */
	@Override
	public void connectionStatusChanged(Kart kart, boolean online) {
		setEnabledForAllSubviews(view, online);

		// Close the activity if the connection to the kart was lost.
		if (!online)
			finish();
	}
	
	@Override
	public void message(Kart kart, String message) {
		// The kart has generated a message, show it as a popup (toast) on the screen.
		Toast.makeText(this, message, Toast.LENGTH_SHORT).show();
	}

	
	/*** OnClickListener implementation ***********************************************************/
	
	@Override
	public void onClick(View view) {
		// Determine which view was clicked.
		switch (view.getId()) {
			// If the used clicked on the faster button, increase the kart speed.
			case R.id.FasterButton:
				kart.increaseDriveSpeed();
				break;
			
			// If the used clicked on the slower button, decrease the kart speed.
			case R.id.SlowerButton:
				kart.decreaseDriveSpeed();
				break;
		}
	}

	
	/*** OnCheckedChangeListener implementation ***************************************************/
	
	@Override
	public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
		// Which check box has been changed?
		switch (buttonView.getId()) {
			// Change output of led 1.
			case R.id.Led1:
				kart.setLed(0, isChecked);
				break;
			
			// Change output of led 2.
			case R.id.Led2:
				kart.setLed(1, isChecked);
				break;
				
			// Change output of led 3.
			case R.id.Led3:
				kart.setLed(2, isChecked);
				break;
			
			// Change output of led 4.
			case R.id.Led4:
				kart.setLed(3, isChecked);
				break;
		}
	}

	
	/*** OnTouchListener implementation ***********************************************************/
	
	@Override
	public boolean onTouch(View view, MotionEvent event) {
		// Which button was touched?
		switch (view.getId()) {
			case R.id.LeftButton:
				// If the left button is actually touched, move the steering to the left.
				if (event.getAction() == MotionEvent.ACTION_DOWN)
					kart.setSteeringPositionNormalized(-1f);
				// If the left button is actually released, move the steering to the center.
				else if (event.getAction() == MotionEvent.ACTION_UP)
					kart.setSteeringPositionNormalized(0);
				return true;
				
			case R.id.RightButton:
				// If the right button is actually touched, move the steering to the right.
				if (event.getAction() == MotionEvent.ACTION_DOWN)
					kart.setSteeringPositionNormalized(1f);
				// If the right button is actually released, move the steering to the center.
				else if (event.getAction() == MotionEvent.ACTION_UP)
					kart.setSteeringPositionNormalized(0);
				return true;
		
			default:
				return false;
		}
	}
}