package ch.hevs.utils.view;

import android.content.Context;
import android.graphics.Canvas;
import android.util.AttributeSet;
import android.view.MotionEvent;
import android.widget.SeekBar;

import ch.hevs.utils.math.Limits;

/**
 * Enables to render a seek bar view vertically.
 * <p>
 * This class does all the needed modifications in order to render the seek bar vertically and
 * modifies its touch behavior accordingly. You can use the class the same way you would use any
 * UI view component.
 *
 * @author Chi Kheang
 * @version 2.1 - 2015
 */
@SuppressWarnings("NullableProblems")
public class VerticalSeekBar extends SeekBar {

    private int width, height, oldWidth, oldHeight;
    private OnSeekBarChangeListener onChangeListener;
    private int lastProgress = 0;

    public VerticalSeekBar(Context context) {
        super(context);
    }

    public VerticalSeekBar(Context context, AttributeSet attributeSet, int defStyle) {
        super(context, attributeSet, defStyle);
    }

    public VerticalSeekBar(Context context, AttributeSet attributeSet) {
        super(context, attributeSet);
    }

    @Override
    protected void onSizeChanged(int width, int height, int oldWidth, int oldHeight) {
        // Call method on superclass, but change width and height.
        super.onSizeChanged(height, width, oldHeight, oldWidth);

        // Save dimensions for later use.
        this.width = width;
        this.height = height;
        this.oldWidth = oldWidth;
        this.oldHeight = oldHeight;
    }

    @Override
    public synchronized void setProgress(int progress) {
        // Set progress on superclass.
        super.setProgress(progress);

        // Readjust size.
        onSizeChanged(width, height, oldWidth, oldHeight);

        // Notify listener about the change.
        if (onChangeListener != null)
            onChangeListener.onProgressChanged(this, progress, false);
    }

    @Override
    protected synchronized void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
        // Call method on superclass with height and width changed.

        //noinspection SuspiciousNameCombination
        super.onMeasure(heightMeasureSpec, widthMeasureSpec);

        // Set the dimensions.
        setMeasuredDimension(getMeasuredHeight(), getMeasuredWidth());
    }

    @Override
    protected void onDraw(Canvas canvas) {
        // Rotate slider 90 degrees clockwise.
        canvas.rotate(-90);

        // Translate slider the height to the left.
        canvas.translate(-getHeight(), 0);

        // Use the drawing implementation of the superclass.
        super.onDraw(canvas);
    }

    @Override
    public void setOnSeekBarChangeListener(OnSeekBarChangeListener onChangeListener) {
        this.onChangeListener = onChangeListener;
    }

    @Override
    public boolean onTouchEvent(MotionEvent event) {
        if (!isEnabled()) {
            return false;
        }

        switch (event.getAction()) {
            case MotionEvent.ACTION_DOWN:
                if (onChangeListener != null)
                    onChangeListener.onStartTrackingTouch(this);
                setPressed(true);
                setSelected(true);
                break;

            case MotionEvent.ACTION_MOVE:
                // Calling the super seems to help fix drawing problems...
                super.onTouchEvent(event);

                // Convert the touch position into a seek bar progress value.
                int max = getMax();
                int progress = Limits.limit(max - (int)(max * event.getY() / getHeight()),
                        0, max);

                // Update progress.
                setProgress(progress);

                // Only inform listener if the progress has actually changed, otherwise the
                // listener gets called ~5 times per change.
                if (progress != lastProgress) {
                    lastProgress = progress;
                    if (onChangeListener != null)
                        onChangeListener.onProgressChanged(this, progress, true);
                }

                onSizeChanged(getWidth(), getHeight(), 0, 0);
                setPressed(true);
                setSelected(true);
                break;

            case MotionEvent.ACTION_UP:
                if (onChangeListener != null)
                    onChangeListener.onStopTrackingTouch(this);
                setPressed(false);
                setSelected(false);
                break;

            case MotionEvent.ACTION_CANCEL:
                super.onTouchEvent(event);
                setPressed(false);
                setSelected(false);
                break;
        }

        return true;
    }
}
